/*
 * ti_file - a class used to handle files for ti calculators
 *
 * Copyright (C) 2010 Michael Jordan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "ti_file.h"

	ti_file::ti_file(string filename) {
		is_inited = false;
		if(is_valid(filename)) {
			ifstream tifile (filename.c_str(), ios::in | ios::binary);
			if(tifile.is_open()) {
				hd.id_str = "";
				hd.entry = NULL;
				unsigned char *buf = new unsigned char [256];
				if(!tifile.read((char *)buf, 8)) {
					// Error reading from file
					delete[] buf;
					tifile.close();
					return;
				}
				buf[8] = 0;
				hd.id_str.append((char *)buf);
				if(hd.id_str.compare("**TIFL**") == 0) {
					// This is a flash file
					hd.entry = new ti_file::flash_head;
					tifile.seekg(0x11);
					if(!tifile.read((char *)buf, 8)) {
						// Error reading from file
						delete[] buf;
						tifile.close();
						return;
					}
					buf[8] = 0;
					flash_head *tmpptr = (flash_head *)hd.entry;
					tmpptr->name.append((char *)buf);
					tifile.seekg(0x8);
					if(!tifile.read((char *)buf, 2)) {
						// Error reading from file
						delete (flash_head *)hd.entry;
						tifile.close();
						delete[] buf;
						return;
					}
					tmpptr->version[0] = buf[0];
					tmpptr->version[1] = buf[1];
					tifile.seekg(0xC);
					if(!tifile.read((char *)buf, 4)) {
						// Error reading from file
						delete (flash_head *)hd.entry;
						tifile.close();
						delete[] buf;
						return;
					}
					tmpptr->bcd_date[0] = buf[0];
					tmpptr->bcd_date[1] = buf[1];
					tmpptr->bcd_date[2] = buf[2];
					tmpptr->bcd_date[3] = buf[3];
					tifile.seekg(0x30);
					if(!tifile.read((char *)buf, 1)) {
						// Error reading from file
						delete (flash_head *)hd.entry;
						tifile.close();
						delete[] buf;
						return;
					}
					tmpptr->dev_type = buf[0];
					tifile.seekg(0x31);
					if(!tifile.read((char *)buf, 1)) {
						// Error reading from file
						delete (flash_head *)hd.entry;
						tifile.close();
						delete[] buf;
						return;
					}
					tmpptr->data_type = buf[0];
					tifile.seekg(0x48);
					if(!tifile.read((char *)buf, 1)) {
						// Error reading from file
						delete (flash_head *)hd.entry;
						tifile.close();
						delete[] buf;
						return;
					}
					tmpptr->hdw_id = buf[0];
					tifile.seekg(0x4A);
					if(!tifile.read((char *)buf, 4)) {
						// Error reading from file
						delete (flash_head *)hd.entry;
						tifile.close();
						delete[] buf;
						return;
					}
					tmpptr->data_size = buf[0];
					tmpptr->data_size += (buf[1] * 0x100);
					tmpptr->data_size += (buf[2] * 0x10000);
					tmpptr->data_size += (buf[3] * 0x1000000);
					tmpptr->data = new unsigned char[tmpptr->data_size];
					if(!tifile.read((char *)(tmpptr->data), tmpptr->data_size)) {
						// Error reading from file
						delete[]tmpptr->data;
						delete (flash_head *)hd.entry;
						tifile.close();
						delete[] buf;
						return;
					}
					delete[] buf;
					tifile.close();
					is_inited = true;
					return;
				} else if((hd.id_str.compare("**TI89**") == 0) ||
						(hd.id_str.compare("**TI92**") == 0) ||
						(hd.id_str.compare("**TI92P*") == 0)) {
					// Parse a variable file
					hd.entry = new ti_file::var_head;
					var_head *tmpptr = (var_head *)hd.entry;
					tifile.seekg(0xA);
					if(!tifile.read((char *)buf, 8)) {
						// Error reading from file
						delete (var_head *)hd.entry;
						tifile.close();
						delete[] buf;
						return;
					}
					buf[8] = 0;
					tmpptr->fld_default.append((char *)buf);
					tifile.seekg(0x12);
					if(!tifile.read((char *)buf, 40)) {
						// Error reading from file
						delete (var_head *)hd.entry;
						tifile.close();
						delete[] buf;
						return;
					}
					buf[40] = 0;
					tmpptr->comment.append((char *)buf);
					tifile.seekg(0x3A);
					if(!tifile.read((char *)buf, 2)) {
						// Error reading from file
						delete (var_head *)hd.entry;
						tifile.close();
						delete[] buf;
						return;
					}
					tmpptr->num_entries = buf[0];
					tmpptr->num_entries += (buf[1] * 0x100);
					int offset = 0x3C;
					tmpptr->entries = new ti_file::vf_entry[tmpptr->num_entries];
					for(int i = 0; i < tmpptr->num_entries; i++) {
						tmpptr->entries[i].parent_folder = tmpptr->fld_default;
						int base = (offset + (i * 0x10));
						tifile.seekg(base);
						if(!tifile.read((char *)buf, 4)) {
							// Error reading from file
							delete tmpptr->entries;
							delete (var_head *)hd.entry;
							tifile.close();
							delete[] buf;
							return;
						}
						tmpptr->entries[i].offset = buf[0];
						tmpptr->entries[i].offset += (buf[1] * 0x100);
						tmpptr->entries[i].offset += (buf[2] * 0x10000);
						tmpptr->entries[i].offset += (buf[3] * 0x1000000);
						tmpptr->entries[i].offset += 4;
						tifile.seekg(base + 4);
						if(!tifile.read((char *)buf, 8)) {
							// Error reading from file
							delete[]tmpptr->entries;
							delete (var_head *)hd.entry;
							tifile.close();
							delete[] buf;
							return;
						}
						buf[8] = 0;
						tmpptr->entries[i].name.append((char *)buf);
						tifile.seekg(base + 12);
						if(!tifile.read((char *)buf, 1)) {
							// Error reading from file
							delete[]tmpptr->entries;
							delete (var_head *)hd.entry;
							tifile.close();
							delete[] buf;
							return;
						}
						tmpptr->entries[i].type_id = buf[0];
						tmpptr->entries[i].type_id += 0xF00C0000;
						tifile.seekg(base + 13);
						if(!tifile.read((char *)buf, 1)) {
							// Error reading from file
							delete[]tmpptr->entries;
							delete (var_head *)hd.entry;
							tifile.close();
							delete[] buf;
							return;
						}
						switch(buf[0]) {
						case 0:
							tmpptr->entries[i].archived = false;
							tmpptr->entries[i].locked = false;
							break;
						case 1:
							tmpptr->entries[i].archived = false;
							tmpptr->entries[i].locked = true;
							break;
						case 2:
							tmpptr->entries[i].archived = true;
							tmpptr->entries[i].locked = false;
							break;
						case 3:
							tmpptr->entries[i].archived = true;
							tmpptr->entries[i].locked = true;
							break;
						}
						tifile.seekg(base + 14);
						if(!tifile.read((char *)buf, 2)) {
							// Error reading from file
							delete[]tmpptr->entries;
							delete (var_head *)hd.entry;
							tifile.close();
							delete[] buf;
							return;
						}
						tmpptr->entries[i].num_entries = buf[0];
						tmpptr->entries[i].num_entries += (buf[1] * 0x100);
					}
					// Go back through and process offsets to get data sections
					// then process the checksums to populate those fields
					struct stat st;
					stat(filename.c_str(), &st);
					for(int i = 0; i < tmpptr->num_entries; i++) {
						if((i + 1) >= tmpptr->num_entries) {
							tmpptr->entries[i].data_len =
									(st.st_size - tmpptr->entries[i].offset) - 2;
						} else {
							if((tmpptr->entries[i].type_id & 0xFF) != 0x1F) {
								tmpptr->entries[i].data_len =
										(tmpptr->entries[i + 1].offset -
										tmpptr->entries[i].offset) - 6;
							} else tmpptr->entries[i].data_len = 0;
						}
						if((tmpptr->entries[i].type_id & 0xFF) == 0x1F) {
							tmpptr->entries[i].parent_folder = "";
							for(int j = 1; j <= tmpptr->entries[i].num_entries; j++) {
								tmpptr->entries[i + j].parent_folder =
										tmpptr->entries[i].name;
							}
						}
					}
					for(int i = 0; i < tmpptr->num_entries; i++) {
						tifile.seekg(tmpptr->entries[i].offset);
						tmpptr->entries[i].data =
								new unsigned char[tmpptr->entries[i].data_len];
						if(!tifile.read((char *)(tmpptr->entries[i].data),
								tmpptr->entries[i].data_len)) {
							// Error reading from file
							for(int j = 0; j <= i; j++)
								delete[]tmpptr->entries[i].data;
							delete tmpptr->entries;
							delete (var_head *)hd.entry;
							tifile.close();
							delete[] buf;
							return;
						}
						unsigned int cs = 0;
						for(int j = 0; j < tmpptr->entries[i].data_len; j++) {
							cs += tmpptr->entries[i].data[j];
						}
						if(!tifile.read((char *)buf, 2)) {
							// Error reading from file
							for(int j = 0; j <= i; j++)
								delete[]tmpptr->entries[i].data;
							delete tmpptr->entries;
							delete (var_head *)hd.entry;
							tifile.close();
							delete[] buf;
							return;
						}
						unsigned int varsum = buf[0];
						varsum += (buf[1] * 0x100);
						tmpptr->entries[i].cksum_ok = ((cs & 0xFFFF) == varsum);
					}
					is_inited = true;
					delete[] buf;
					return;
				}
			}
		}
	}

	ti_file::~ti_file() {
		if(is_good()) {
			is_inited = false;
			if(hd.id_str.compare("**TIFL**") == 0) {
				flash_head *tmpptr = (flash_head *)hd.entry;
				delete[]tmpptr->data;
				delete (flash_head *)hd.entry;
				return;
			}
			var_head *tmpptr = (var_head *)hd.entry;
			for(int i = 0; i < tmpptr->num_entries; i++) delete[]tmpptr->entries[i].data;
			delete[] tmpptr->entries;
			delete (var_head *)hd.entry;
			return;
		}
		return;
	}

	bool ti_file::is_flash() {
		if(is_inited) {
			if(hd.id_str.compare("**TIFL**") == 0) return true;
			else return false;
		}
		return false;
	}

	bool ti_file::is_var() {
		if(is_inited) {
			if((hd.id_str.compare("**TI92**") == 0) || (hd.id_str.compare("**TI89**") ||
					hd.id_str.compare("**TI92P*"))) return true;
			else return false;
		}
		return false;
	}

	int ti_file::num_entries() {
		if(is_inited) {
			if(is_flash()) {
				return 1;
			}
			var_head *tmpptr;
			tmpptr = (var_head *)hd.entry;
			int i = tmpptr->num_entries;
			return i;
		}
		else return -1;
	}

	int ti_file::data_size(int ent_no) {
		if(is_inited) {
			if(is_flash()) {
				if(ent_no != 0) return -1;
				flash_head *tmpptr = (flash_head *)hd.entry;
				int i = tmpptr->data_size;
				return i;
			}
			var_head *tmpptr = (var_head *)hd.entry;
			if((ent_no < tmpptr->num_entries) && !(ent_no < 0))
				return tmpptr->entries[ent_no].data_len;
			return -1;
		}
		return -1;
	}

	bool ti_file::cs_ok(int ent_no) {
		if(is_inited) {
			if(is_flash()) {
				return true;
			}
			var_head *tmpptr = (var_head *)hd.entry;
			if((ent_no < tmpptr->num_entries) && !(ent_no < 0))
				return tmpptr->entries[ent_no].cksum_ok;
			return false;
		}
		return false;
	}

	bool ti_file::is_good() {
		return is_inited;
	}

	bool ti_file::data(int ent_no, unsigned char *buf) {
		if(is_inited) {
			if(is_flash()) {
				if(ent_no != 0) return false;
				flash_head *tmpptr = (flash_head *)hd.entry;
				for(int i = 0; i < tmpptr->data_size; i++)
					buf[i] = tmpptr->data[i];
				return true;
			}
			var_head *tmpptr = (var_head *)hd.entry;
			if((ent_no < tmpptr->num_entries) && !(ent_no < 0)) {
				for(int i = 0; i < tmpptr->entries[ent_no].data_len; i++)
					buf[i] = tmpptr->entries[ent_no].data[i];
				return true;
			} else return false;
		}
		return false;
	}

	string ti_file::name(int ent_no) {
		if(is_inited) {
			if(is_flash()) {
				if(ent_no != 0) return "";
				flash_head *tmpptr = (flash_head *)hd.entry;
				return tmpptr->name;
			}
			var_head *tmpptr = (var_head *)hd.entry;
			if((ent_no < tmpptr->num_entries) && !(ent_no < 0))
				return tmpptr->entries[ent_no].name;
			return "";
		}
		return "";
	}

	bool ti_file::archived(int ent_no) {
		if(is_inited) {
			if(is_flash()) return true;
			var_head *tmpptr = (var_head *)hd.entry;
			if((ent_no < tmpptr->num_entries) && !(ent_no < 0))
				return tmpptr->entries[ent_no].archived;
		}
		return false;
	}

	bool ti_file::locked(int ent_no) {
		if(is_inited) {
			if(is_flash()) return true;
			var_head *tmpptr = (var_head *)hd.entry;
			if((ent_no < tmpptr->num_entries) && !(ent_no < 0))
				return tmpptr->entries[ent_no].locked;
		}
		return false;
	}

	string ti_file::comment() {
		if(is_inited) {
			if(is_flash()) return "";
			var_head *tmpptr = (var_head *)hd.entry;
			return tmpptr->comment;
		}
		return "";
	}

	string ti_file::date() {
		if(is_inited) {
			if(is_flash()) {
				flash_head *tmpptr = (flash_head *) hd.entry;
				string date = "";
				stringstream day;
				day.fill('0');
				day.width(2);
				day << hex << (int)tmpptr->bcd_date[0];
				date = date + day.str() + "/";
				stringstream month;
				month.fill('0');
				month.width(2);
				month << hex << (int)tmpptr->bcd_date[1];
				date = date + month.str() + "/";
				stringstream year_hi;
				year_hi.fill('0');
				year_hi.width(2);
				year_hi << hex << (int)tmpptr->bcd_date[2];
				date = date + year_hi.str();
				stringstream year_lo;
				year_lo.fill('0');
				year_lo.width(2);
				year_lo << hex << (int)tmpptr->bcd_date[3];
				date = date + year_lo.str();
				return date;
			}
			return "";
		}
		return "";
	}

	unsigned int ti_file::type_id(int ent_no) {
		if(is_inited) {
			if(is_flash()) {
				if(ent_no == 0) {
					flash_head *tmpptr = (flash_head *)hd.entry;
					return (tmpptr->data_type + 0xF00C0000);
				}
				return -1;
			}
			var_head *tmpptr = (var_head *)hd.entry;
			if((ent_no < tmpptr->num_entries) && !(ent_no < 0))
				return tmpptr->entries[ent_no].type_id;
			return -1;
		}
		return -1;
	}

	int ti_file::dev_type() {
		if(is_inited) {
			if(is_flash()) {
				flash_head *tmpptr = (flash_head *)hd.entry;
				return tmpptr->dev_type;
			}
			return -1;
		}
		return -1;
	}

	int ti_file::data_type() {
		if(is_inited) {
			if(is_flash()) {
				flash_head *tmpptr = (flash_head *)hd.entry;
				return tmpptr->data_type;
			}
			return -1;
		}
		return -1;
	}

	int ti_file::hdw_id() {
		if(is_inited) {
			if(is_flash()) {
				flash_head *tmpptr = (flash_head *)hd.entry;
				return tmpptr->hdw_id;
			}
			return -1;
		}
		return -1;
	}

	string ti_file::folder(int ent_no) {
		if(is_inited) {
			if(is_flash()) return "";
			var_head *tmpptr = (var_head *)hd.entry;
			if((ent_no < tmpptr->num_entries) && !(ent_no < 0))
				return tmpptr->entries[ent_no].parent_folder;
			return "";
		}
		return "";
	}

	string ti_file::version() {
		if(is_inited) {
			if(is_flash()) {
				flash_head *tmpptr = (flash_head *)hd.entry;
				stringstream version_hi, version_lo;
				string rev;
				version_hi.fill('0');
				version_lo.fill('0');
				version_hi.width(2);
				version_lo.width(2);
				version_hi << hex << (int)tmpptr->version[0];
				version_lo << hex << (int)tmpptr->version[1];
				rev = rev + version_hi.str() + "." + version_lo.str();
				return rev;
			}
			return "";
		}
		return "";
	}

	bool ti_file::is_valid(string filename) {
		struct stat st;
		bool exist, is_dir;

		exist = (stat(filename.c_str(), &st) == 0);
		is_dir = (S_ISDIR(st.st_mode));
		if(exist && !is_dir && st.st_size > 82) return true;
		return false;
	}

	int ti_file::get_sys_directory(string dir, vector<std::string> &files) {
	    DIR *dp;
	    struct dirent *dirp;
	    if((dp  = opendir(dir.c_str())) == NULL) {
	        return errno;
	    }

	    while ((dirp = readdir(dp)) != NULL) {
	        files.push_back(string(dirp->d_name));
	    }
	    closedir(dp);
	    return 0;
	}

	string ti_file::typeid_to_string(int type_id) {
		switch(type_id & 0xFF) {
		case 0x00:
			return "EXPR";
			break;
		case 0x04:
			return "FIG";
			break;
		case 0x0C:
			return "STR";
			break;
		case 0x0E:
			return "LIST";
			break;
		case 0x10:
			return "PICT";
			break;
		case 0x12:
			return "PRGM";
			break;
		case 0x13:
			return "FUNCT";
			break;
		case 0x1C:
			return "AppVar";
			break;
		case 0x1F:
			return "Folder";
			break;
		case 0x21:
			return "ASM";
			break;
		case 0x23:
			return "OS";
			break;
		case 0x24:
			return "Flash APP";
			break;
		case 0x25:
			return "Certificate";
			break;
		case 0x3E:
			return "License";
			break;
		default:
			return "Unknown";
			break;
		}
	}
